/*******************************************************************************
 *
 * \file
 * \brief Implement utilities for source code locations
 *
 * Authors: Maxime Arthaud
 *
 * Contact: ikos@lists.nasa.gov
 *
 * Notices:
 *
 * Copyright (c) 2018-2019 United States Government as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All Rights Reserved.
 *
 * Disclaimers:
 *
 * No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF
 * ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED
 * TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO SPECIFICATIONS,
 * ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE,
 * OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
 * ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO
 * THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN
 * ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR RECIPIENT OF ANY RESULTS,
 * RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY OTHER APPLICATIONS
 * RESULTING FROM USE OF THE SUBJECT SOFTWARE.  FURTHER, GOVERNMENT AGENCY
 * DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE,
 * IF PRESENT IN THE ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
 *
 * Waiver and Indemnity:  RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
 * THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL
 * AS ANY PRIOR RECIPIENT.  IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS
 * IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH
 * USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
 * RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD
 * HARMLESS THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS,
 * AS WELL AS ANY PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW.
 * RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
 * UNILATERAL TERMINATION OF THIS AGREEMENT.
 *
 ******************************************************************************/

#include <ikos/analyzer/util/source_location.hpp>

namespace ikos {
namespace analyzer {

/// \brief Return the relative path between `from` and `to`
static boost::filesystem::path relative_to(const boost::filesystem::path& from,
                                           const boost::filesystem::path& to) {
  auto from_it = from.begin();
  auto to_it = to.begin();

  // loop through both
  while (from_it != from.end() && to_it != to.end() && *to_it == *from_it) {
    ++to_it;
    ++from_it;
  }

  boost::filesystem::path final_path;
  while (from_it != from.end()) {
    final_path /= "..";
    ++from_it;
  }

  while (to_it != to.end()) {
    final_path /= *to_it;
    ++to_it;
  }

  return final_path;
}

std::string source_location_string(SourceLocation loc,
                                   const boost::filesystem::path& wd) {
  if (!loc) {
    return "?:?:?";
  }

  boost::filesystem::path abs_path = loc.path();
  boost::filesystem::path rel_path = relative_to(wd, abs_path);
  boost::filesystem::path pretty_path =
      (rel_path.native().size() < abs_path.native().size()) ? rel_path
                                                            : abs_path;
  std::string r;
  r += pretty_path.string();
  r += ':';
  r += std::to_string(loc.line());
  r += ':';
  r += std::to_string(loc.column());
  return r;
}

} // end namespace analyzer
} // end namespace ikos
